/*
 * USB CONSOLE over host-to-host cable.
 *
 * Copyright (c) 2003          Daniele Bellucci <bellucda@tiscali.it>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/console.h>
#include <linux/completion.h>
#include <linux/usb.h>

#define DRIVER_VERSION "v1.0"
#define DRIVER_AUTHOR  "Daniele Bellucci <bellucda@tiscali.it>"
#define DRIVER_DESC    "USB console over host-to-host cable"

#define USB_CONSOLE_VENDOR_ID	0x067b  /* PL2301 vendorId  */
#define USB_CONSOLE_PRODUCT_ID	0x0000  /* PL2301 productId */


static struct usb_device_id usbconsole_table [] = {
	{ USB_DEVICE(USB_CONSOLE_VENDOR_ID, USB_CONSOLE_PRODUCT_ID) },
	{ }
};


MODULE_DEVICE_TABLE(usb, usbconsole_table);

static int 
usbconsole_probe(struct usb_interface *, const struct usb_device_id *);

static void 
usbconsole_disconnect(struct usb_interface *);

static void 
usbconsole_write(struct console *, const char *, unsigned );

static void 
usbconsole_write_bulk_callback(struct urb *urb, struct pt_regs *regs);


struct usbconsole_dev {
	struct usb_device *    udev;		  /* usb device pointer   */
	struct usb_interface * interface;	  /* usb device interface */

	__u8		       out_endpointAddr;  /* bulk out endpoint addr. */
	size_t                 out_buffer_size;
	atomic_t	       out_busy;	  /* prevent races between 
						   * urb completion handler and
						   * the disconnect routine */

	struct completion      out_finished;	  /* wait for the write 
						   * callback to finish */

	unsigned int           out_pipe;          /* pipe to send bulk msg*/
};





static struct console usbconsole = {
	.name     = "usbcons",
	.write    = usbconsole_write,
	.flags    = CON_PRINTBUFFER,
	.index    = -1,
};



static void 
usbconsole_write(struct console *console, const char *buff, unsigned count)
{
	struct usbconsole_dev *dev = (struct usbconsole_dev *)console->data;
	struct urb *out_urb;

	unsigned int index = 0;
	unsigned char *out_buffer;
	unsigned length;

	while (count) {
		atomic_inc(&dev->out_busy);
		out_urb = usb_alloc_urb(0, GFP_ATOMIC);

		if (out_urb == NULL) {
			atomic_dec(&dev->out_busy);
			return;
		}

		out_urb->transfer_flags = (URB_NO_TRANSFER_DMA_MAP |
						     URB_ASYNC_UNLINK);
			
		length = min(dev->out_buffer_size, count);
		out_buffer = usb_buffer_alloc(dev->udev, length, GFP_ATOMIC,
					      &out_urb->transfer_dma);
			
		if (out_buffer == NULL) {
			atomic_dec(&dev->out_busy);
			return;
		}

		memcpy(out_buffer, buff+index, length);
		usb_fill_bulk_urb(out_urb, dev->udev, 
				  dev->out_pipe,
				  out_buffer, length, 
				  usbconsole_write_bulk_callback,
				  dev);
		
		usb_submit_urb(out_urb, GFP_ATOMIC);
		count -= length;
		index += length;
	}
}




static struct usb_driver usbconsole_driver = {
	.owner      =	THIS_MODULE,
	.name       =	"usbconsole",
	.probe      =	usbconsole_probe,
	.disconnect =	usbconsole_disconnect,
	.id_table   =	usbconsole_table,
};




static void usbconsole_write_bulk_callback(struct urb *urb, struct pt_regs *regs)
{

	struct usbconsole_dev *dev;
	dev = (struct usbconsole_dev *)urb->context;

	usb_buffer_free(urb->dev, urb->transfer_buffer_length,
			urb->transfer_buffer, urb->transfer_dma);

	usb_free_urb(urb);
	if (atomic_dec_and_test(&dev->out_busy))
		complete(&dev->out_finished);
}




static int usbconsole_probe(struct usb_interface *interface, 
			    const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(interface);
	struct usbconsole_dev *dev = NULL;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	int i, retval = -ENOMEM;

	/* allocate memory for our device state and initialize it */
	dev = kmalloc(sizeof(struct usbconsole_dev), GFP_KERNEL);
	if (dev == NULL) {
		err("Out of memory");
		return -ENOMEM;
	}
	memset(dev, 0, sizeof(*dev));

	dev->udev = udev;
	dev->interface = interface;
	iface_desc = &interface->altsetting[0];

	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;
		if ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) !=
		    USB_ENDPOINT_XFER_BULK)
			continue;

		/* .. skip bulk in endpoint .. */
		if (endpoint->bEndpointAddress & USB_DIR_IN)
			continue;

		dev->out_endpointAddr = endpoint->bEndpointAddress;
		dev->out_buffer_size = endpoint->wMaxPacketSize;
		dev->out_pipe = usb_sndbulkpipe(udev, 
						endpoint->bEndpointAddress);
	}
	
	/* check if bulk-out endpoint has been found */
	if ( dev->out_endpointAddr == 0) {
		retval = -ENODEV;
		kfree(dev);
		err("Couldn't find bulk-out endpoint");
		return retval;
	}


	usb_set_intfdata(interface, dev);
	init_completion (&dev->out_finished);
	usbconsole.data = dev;
	register_console(&usbconsole);

	return 0;
}


static void usbconsole_disconnect(struct usb_interface *interface)
{
	struct usbconsole_dev *dev;

	unregister_console(&usbconsole);
	
	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);

	/* wait for ongoing write completion */
	if (atomic_read(&dev->out_busy))
		wait_for_completion(&dev->out_finished);
}




static int __init usbconsole_init(void)
{
	int retval;
	retval = usb_register(&usbconsole_driver);
	if (retval)
		info(DRIVER_VERSION " : " DRIVER_DESC);
	return retval;
}


static void __exit usbconsole_exit(void)
{
	usb_deregister(&usbconsole_driver);
}


module_init(usbconsole_init);
module_exit(usbconsole_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
